import time, types, re, sys
import Pyro.core, Pyro.naming, Pyro.util, Pyro.constants
from Pyro.errors import *

Log=Pyro.util.Log

#if not Pyro.util.supports_multithreading():
#	raise NotImplementedError('Event Server needs Python threads support.')


# EVENT - the thing that is published. Has a subject and contains a message.
class Event:
	def __init__(self, subject, msg, creationTime=None):
		self.msg=msg
		self.subject=subject
		self.time=creationTime or time.time()
	def __str__(self):
		return "<EVENT SUBJ %s (%s): %s>" % (self.subject, time.ctime(self.time), str(self.msg))


# The EVENTSERVICE is the actual Pyro server.
# The data structures are protected by a Semaphore lock,
# to prevent several threads accessing them at the same time.
# Because the callback method ('event') is set to ONEWAY mode,
# there are no deadlock issues. They would otherwise occur
# when a client calls back to the ES while still processing the
# event callback.
#
# BTW: Subscribers are remembered trough their proxy class.
# This class is capable of being a correct key in a dictionary.
class EventService(Pyro.core.ObjBase):
	def __init__(self):
		Pyro.core.ObjBase.__init__(self)
		self.subscribers={}			# subject -> { subscriber -> dummy }
		self.subscribersMatch={}	# subjectPattern -> { subscriber -> dummy }
		self.lock=Pyro.util.getLockObject()
	def _mksequence(self, seq):
		if not (type(seq) in (types.TupleType,types.ListType)):
			seq=(seq,)
		return seq
	def subscribe(self, subjects, subscriber):
		if not subjects: return
		self.lock.acquire()
		subscriber._setOneway('event') # don't wait for answer for this method
		# Subscribe into a dictionary; this way; somebody can subscribe
		# only once to this subject. Subjects are exact strings.
		for subject in self._mksequence(subjects):
			self.subscribers.setdefault(subject.lower(),{}) [subscriber]=1
		self.lock.release()	
	def subscribeMatch(self, subjects, subscriber):
		if not subjects: return
		self.lock.acquire()
		subscriber._setOneway('event') # don't wait for answer for this method
		# Subscribe into a dictionary; this way; somebody can subscribe
		# only once to this subject. Subjects are regex patterns.
		for subject in self._mksequence(subjects):
			matcher = re.compile(subject,re.IGNORECASE)
			self.subscribersMatch.setdefault(matcher,{}) [subscriber]=1
		self.lock.release()	
	def unsubscribe(self, subjects, subscriber):
		if not subjects: return
		self.lock.acquire()
		for subject in self._mksequence(subjects):
			try:
				blaat=self.subscribers[subject.lower()]
				del self.subscribers[subject.lower()] [subscriber]
			except KeyError:
				try:
					m=re.compile(subject,re.IGNORECASE)
					del self.subscribersMatch[m] [subscriber]
				except KeyError:
					pass

		self.lock.release()			
	def publish(self, subjects, message):
		if not subjects: return
		# keep the creation time, this must be the same for all events.
		creationTime=time.time()
		# now enter the lock.
		self.lock.acquire()
		# publish a message. Subjects must be exact strings
		for subject in self._mksequence(subjects):
			event = Event(subject, message, creationTime)
			subjectLC=subject.lower()
			try:
				for s in self.subscribers[subjectLC].keys():
					try:
						s.event(event) # this is a oneway invocation, no result
					except ConnectionClosedError:
						# hmm, subscriber no longer connected...
						del self.subscribers[subjectLC] [s]
					except:
						# ignore all other errors
						pass
			except KeyError:
				pass
			# process the subject patterns
			for (m,subs) in self.subscribersMatch.items():
				if m.match(subject):
					# send event to all subscribers
					for s in subs.keys():	
						try:
							s.event(event)
						except ConnectionClosedError:
							# hmm, subscriber no longer connected...
							del self.subscribersMatch[m] [s]
						except:
							# ignore all other errors
							pass
		self.lock.release()			



class EventServiceStarter:
	def __init__(self, identification=None):
		Pyro.core.initServer()
		self.running=1
		self.identification=identification
	def start(self,host=None,port=None):
		daemon = Pyro.core.Daemon(host=host,port=port)
		if self.identification:
			daemon.setAllowedIdentifications([self.identification])
			print 'Requiring connection authentication.'

		locator = Pyro.naming.NameServerLocator(identification=self.identification)
		print 'Looking for Name Server...'
		ns = locator.getNS()

		# check if ES already running
		try:
			ns.resolve(Pyro.constants.EVENTSERVER_NAME)
			print 'The Event Server appears to be already running.'
			print 'You cannot start multiple Event Servers.'
			ans=raw_input('Start new Event Server anyway (y/n)? ')
			if ans!='y':
				return
			ns.unregister(Pyro.constants.EVENTSERVER_NAME)
		except NamingError:
			pass

		daemon.useNameServer(ns)

		es = EventService()

		esURI=daemon.connect(es, Pyro.constants.EVENTSERVER_NAME)
		print 'URI=',esURI
		print 'Event Server started.'
		Log.msg('ES daemon','This is the Pyro Event Server.')
		
		# I use a timeout here otherwise you can't break gracefully on Windoze
		try:
			daemon.setTimeout(20)  # XXX fixed timeout
			daemon.requestLoop(lambda s=self: s.running)
		except KeyboardInterrupt:
			Log.warn('ES daemon','shutdown on user break signal')
			print 'Shutting down on user break signal.'
			self.shutdown(es)
		except:
			try:
				import traceback
				(exc_type, exc_value, exc_trb) = sys.exc_info()
				out = ''.join(traceback.format_exception(exc_type, exc_value, exc_trb)[-5:])
				Log.error('ES daemon', 'Unexpected exception, type',exc_type,
					'\n--- partial traceback of this exception follows:\n',
					out,'\n--- end of traceback')
				print '*** Exception occured!!! Partial traceback:'
				print out
				print '*** Resuming operations...'
			finally:	
				del exc_type, exc_value, exc_trb

		Log.msg('ES daemon','Shut down gracefully.')
		print 'Event Server gracefully stopped.'

	def mustContinueRunning(self):
		return self.running
	def shutdown(self,es):
		daemon=es.getDaemon()
		try:
			daemon.disconnect(es) # clean up nicely
		except NamingError,x:
			Log.warn('ES daemon','disconnect error during shutdown:',x)
		except ConnectionClosedError,x:
			Log.warn('ES daemon','lost connection with Name Server, cannot unregister')
		self.running=0
		daemon.shutdown()


def start():
	if __name__!='__main__':
		del sys.argv[0]	# remove '-c' arg
	Args = Pyro.util.ArgParser()
	Args.parse(sys.argv,'hn:p:i:')
	try:
		Args.getOpt('h')
		print 'Usage: es [-h] [-n hostname] [-p port] [-i identification]'
		print '  where -p = ES server port'
		print '        -n = non-default hostname to bind on'
		print '        -i = the required authentication ID for ES clients,'
		print '             also used to connect to other Pyro services'
		print '        -h = print this help'
		raise SystemExit
	except KeyError:
		pass
	host = Args.getOpt('n','')
	port = Args.getOpt('p',None)
	ident = Args.getOpt('i',None)
	if port:
		port=int(port)
	Args.printIgnored()
	if Args.args:
		print 'Ignored arguments:',' '.join(Args.args)

	print '*** Pyro Event Server ***'
	starter=EventServiceStarter(identification=ident)
	starter.start(host,port)

