#############################################################################
#
#	naming.py - Pyro Name Server
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################


import sys, os, socket, string
import dircache, shutil, SocketServer
import Pyro
import util, errors, core, protocol, constants


if os.name!='java':			# Jython has no errno/select modules
	import errno
	import select
else:
	# fake an errno module
	class Errno: pass
	errno=Errno()
	errno.EEXIST=-1
	errno.EBUSY=-1
	errno.ENOENT=-1
	errno.EISDIR=-1
	errno.ENOTDIR=-1
	

Log = util.Log

#############################################################################
#
# The Pyro NameServer Locator.
# Use a broadcast mechanism to find the broadcast server of the NS which
# can provide us with the URI of the NS.
# Can also perform direct lookup (no broadcast) if the host is specified.
# (in that case, the 'port' argument is the Pyro port, not a broadcast port).
#
#############################################################################

class NameServerLocator:
	def __init__(self, identification=None):
		self.identification=identification

	def sendSysCommand(self,request,host=None,port=None,trace=0):
		if not host:
			host=Pyro.config.PYRO_NS_HOSTNAME
		if not port:
			if not host:
				# select the default broadcast port
				port = Pyro.config.PYRO_NS_BC_PORT
			else:
				# select the default port
				port = Pyro.config.PYRO_NS_PORT
		# We must discover the location of the name server.
		# Pyro's NS can answer to broadcast requests.
		try:
			s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
			if not host:
				# No host specified. Use broadcast mechanism
				if os.name=='java':
					# Jython handling for broadcast addrs
					destination = ('255.255.255.255', port)
				else:
					# regular python broadcast socket
					destination = ('<broadcast>', port)
				if hasattr(socket,'SO_BROADCAST'):
					s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
			else:
				# use direct lookup with PYROLOC: mechanism, no broadcast
				if trace:
					print 'Locator: contacting Pyro Name Server...'
				uri=core.PyroURI(host,constants.NAMESERVER_NAME,port,'PYROLOC')
				prox=core.getProxyForURI(uri)
				prox._setIdentification(self.identification)
				prox.ping()	# force resolving of PYROLOC: uri
				return prox.URI # return resolved uri
			
			if trace:
				print 'Locator: searching Pyro Name Server...'
			for i in range(Pyro.config.PYRO_BC_RETRIES+1):
				s.sendto(request, destination)  # send to Pyro NS's bc port
				if os.name!='java':
					ins,outs,exs = select.select([s],[],[s],Pyro.config.PYRO_BC_TIMEOUT)
				else:
					ins=[s]		# XXX preliminary Jython solution (no select)
				if s in ins:
					reply, fromaddr = s.recvfrom(1000)
					return reply
				if trace and i<Pyro.config.PYRO_BC_RETRIES:
					print 'Locator: retry',i+1
		except socket.error,e:
			Log.error('NameServerLocator','network error:',e)
			if trace:
				print 'Locator: network error:',e

		Log.error('NameServerLocator','Name Server not responding to broadcast')
		raise errors.PyroError('Name Server not responding')
	
	def detectNS(self,host=None,port=None,trace=0):
		# just try to detect an existing NS.
		return core.PyroURI(self.sendSysCommand('location',host,port,trace))

	def getNS(self,host=None,port=None,trace=0):
		reply = self.sendSysCommand('location',host,port,trace)
		Log.msg('NameServerLocator','Name Server found:',reply)
		ns=NameServerProxy(core.PyroURI(reply),self.identification)
		ns._setIdentification(self.identification)
		return ns


# NOTE: The NameServerProxy class below is hand crafted.
# This is because we want to enforce the default group name on all name
# arguments that are not absolute.
# In doing so, we make sure that each object name is passed as an
# absolute name (from the name space root) too. This is needed because
# the NS has no state and we can't really 'cd' to a group in the NS.

class NameServerProxy:  
	def __init__(self,URI,identification=None):
		self.URI = URI
		self.objectID = URI.objectID
		self.adapter = protocol.getProtocolAdapter(self.URI.protocol)
		self.adapter.setIdentification(identification)
		self.adapter.bindToURI(URI)
	def ping(self):
		return self.adapter.remoteInvocation('ping',0)
	def register(self,name,URI):
		return self.adapter.remoteInvocation('register',0,self.__expand(name),URI)
	def resolve(self,name):
		return self.adapter.remoteInvocation('resolve',0,self.__expand(name))
	def flatlist(self):
		return self.adapter.remoteInvocation('flatlist',0)
	def unregister(self,name):
		return self.adapter.remoteInvocation('unregister',0,self.__expand(name))
	def createGroup(self,gname):
		return self.adapter.remoteInvocation('createGroup',0,self.__expand(gname))
	def deleteGroup(self,gname):
		return self.adapter.remoteInvocation('deleteGroup',0,self.__expand(gname))
	def list(self,gname):
		return self.adapter.remoteInvocation('list',0,self.__expand(gname))
	def fullName(self,name):
		return self.__expand(name)
	def __expand(self,name):
		if Pyro.config.PYRO_NS_DEFAULTGROUP[0]!=':':
			raise errors.NamingError('default group name is not absolute')
		if name:
			if name[0]==':':
				return name
			return Pyro.config.PYRO_NS_DEFAULTGROUP+'.'+name
		else:
			return Pyro.config.PYRO_NS_DEFAULTGROUP
	def _setIdentification(self, ident):
		self.adapter.setIdentification(ident)



#############################################################################
#
#	The Name Server (a Pyro Object).
#
#	It has more methods than the ones available in the proxy (above)
#	but that is because the other methods are private and for internal
#	use only.
#
#############################################################################


class NameServer(core.ObjBase):
	def __init__(self):
		core.ObjBase.__init__(self)
		self.root=NamedTree('<root>')
		self.lock=util.getLockObject()
		# create default groups
		self.createGroup(':'+'Pyro')
		self.createGroup(Pyro.config.PYRO_NS_DEFAULTGROUP)

	def register(self,name,URI):
		self.validateName(name)
		URI=self.validateURI(URI)
		try:
			self.lock.acquire()
			(group, name)=self.locateGrpAndName(name)
			if isinstance(group,NameValue):
				raise errors.NamingError('parent is no group')
			try:
				group.newleaf(name,URI)
				Log.msg('NameServer','registered',name,'with URI',str(URI))
			except KeyError:
				Log.msg('NameServer','name already exists:',name)
				raise errors.NamingError('name already exists')
		finally:
			self.lock.release()

	def unregister(self,name):
		self.validateName(name)
		try:
			self.lock.acquire()
			(group, name)=self.locateGrpAndName(name)
			try:
				group.cutleaf(name)
				Log.msg('NameServer','unregistered',name)
			except KeyError:
				raise errors.NamingError('name not found')
			except ValueError:
				Log.msg('NameServer','attempt to remove a group:',name)
				raise errors.NamingError('is a group')
		finally:
			self.lock.release()

	def resolve(self,name):
		# not thread-locked: higher performance and not necessary.
		self.validateName(name)
		try:
			branch=self.getBranch(name)
			if isinstance(branch,NameValue):
				return branch.value
			else:
				Log.msg('NameServer','attempt to resolve groupname:',name)
				raise errors.NamingError('attempt to resolve groupname')
		except KeyError:
			raise errors.NamingError('name not found')
		except AttributeError:
			raise errors.NamingError('group not found')

	def flatlist(self):
		# return a dump
		try:
			self.lock.acquire()
			r=self.root.flatten()
		finally:
			self.lock.release()
		for i in range(len(r)):
			r[i]=(':'+r[i][0], r[i][1])
		return r

	def ping(self):
		# Just accept a remote invocation.
		# This method is used to check if NS is still running,
		# and also by the locator if a direct lookup is needed.
		pass

	# --- hierarchical naming support
	def createGroup(self,groupname):
		self.validateName(groupname)
		try:
			self.lock.acquire()
			(parent,name)=self.locateGrpAndName(groupname)
			try:
				parent.newbranch(name)
				Log.msg('NameServer','created group',groupname)
			except KeyError,x:
				raise errors.NamingError(x)
		finally:
			self.lock.release()

	def deleteGroup(self,groupname):
		self.validateName(groupname)
		if groupname==':':
			Log.msg('NameServer','attempt to deleteGroup root group')
			raise errors.NamingError('not allowed to delete root group')
		try:
			self.lock.acquire()
			(parent,name)=self.locateGrpAndName(groupname)
			try:
				parent.cutbranch(name)
				Log.msg('NameServer','deleted group',name)
			except KeyError:
				raise errors.NamingError('group not found')
			except ValueError:
				Log.msg('NameServer','attempt to deleteGroup a non-group:',groupname)
				raise errors.NamingError('is no group')
		finally:
			self.lock.release()
			
	def list(self,groupname):
		# not thread-locked: higher performance and not necessary.
		if not groupname:
			groupname=':'
		self.validateName(groupname)
		try:
			return self.getBranch(groupname).list()
		except KeyError:
			raise errors.NamingError('group not found')
		except AttributeError:
			Log.msg('NameServer','attempt to list a non-group:',groupname)
			raise errors.NamingError('is no group')
			

	# --- private methods follow
	def locateGrpAndName(self,name):
		# ASSUME name is absolute (from root) (which is required here)
		idx=string.rfind(name,'.')
		if idx>=0:
			# name is hierarchical
			grpname=name[:idx]
			name=name[idx+1:]
			try:
				return (self.getBranch(grpname), name)
			except KeyError:
				raise errors.NamingError('(parent)group not found')
		else:
			# name is in root
			return (self.root, name[1:])

	def getBranch(self,name):
		# ASSUME name is absolute (from root) (which is required here)
		name=name[1:]
		if name:
			return reduce(lambda x,y: x[y], string.split(name,'.'), self.root)
		else:
			return self.root

	def validateName(self,name):
		if len(name)>=1 and name[0]==':':
			if ('' not in string.split(name,'.')):
				for i in name:
					if ord(i)<33 or ord(i)>126:
						raise errors.NamingError('invalid name')
				return
			else:
				raise errors.NamingError('invalid name')
		else:
			raise errors.NamingError('name is not absolute')

	def validateURI(self,URI):
		if isinstance(URI, core.PyroURI):
			return URI
		try:
			u = core.PyroURI('')
			u.reinitFromString(URI)
			return u
		except:
			raise errors.NamingError('invalid URI')

	def publishURI(self, uri):
		uri=str(uri)
		print 'URI is:',uri
		try:
			f=open(Pyro.config.PYRO_NS_URIFILE,'w')
			f.write(uri+'\n'); f.close()
			print 'URI written to:',Pyro.config.PYRO_NS_URIFILE
			Log.msg('NameServer','URI written to',Pyro.config.PYRO_NS_URIFILE)
		except:
			Log.warn('NameServer','Couldn\'t write URI to',Pyro.config.PYRO_NS_URIFILE)

#############################################################################
#
#	NamedTree data type. Used for the hierarchical name server.
#
#############################################################################
class NameValue:
	def __init__(self, name,value=None):
		self.name=name
		self.value=value

class NamedTree:
	def __init__(self, name):
		self.name=name
		self.branches={}
	def newbranch(self,name):
		if name in self.branches.keys():
			raise KeyError,'duplicate name'
		t = NamedTree(name)
		self.branches[name]=t
		return t
	def newleaf(self,name,value=None):
		if name in self.branches.keys():
			raise KeyError,'duplicate name'
		l = NameValue(name,value)
		self.branches[name]=l
		return l
	def cutleaf(self,name):
		if isinstance(self.branches[name], NameValue):
			del self.branches[name]
		else:
			raise ValueError,'not a leaf'
	def cutbranch(self,name):
		if isinstance(self.branches[name], NamedTree):
			del self.branches[name]
		else:
			raise ValueError,'not a branch'
	def __getitem__(self,name):
		return self.branches[name]
	def list(self):
		l=[]
		for (k,v) in self.branches.items():
			if isinstance(v, NamedTree):
				l.append( (k,0) )	# tree
			elif isinstance(v, NameValue):
				l.append( (k,1) )	# leaf
			else:
				raise ValueError('corrupt tree')
		return l
	def flatten(self,prefix=''):
		flat=[]
		for (k,v) in self.branches.items():
			if isinstance(v, NameValue):
				flat.append( (prefix+k, v.value) )
			elif isinstance(v, NamedTree):
				flat.extend(v.flatten(prefix+k+'.'))
		return flat
				

		
#############################################################################
#
#	The Persistent Name Server (a Pyro Object).
#	This implementation uses the hierarchical file system to
#	store the groups (as directories) and objects (as files).
#
#############################################################################

class PersistentNameServer(NameServer):
	def __init__(self, dbdir=None):
		# root is not a NamedTree but a directory
		self.dbroot=os.path.join(Pyro.config.PYRO_STORAGE,dbdir or 'Pyro_NS_database')
		try:
			os.mkdir(self.dbroot)
		except OSError,x:
			if x.errno not in (errno.EEXIST, errno.EBUSY):
				raise
		try:
			NameServer.__init__(self)
		except errors.NamingError:
			pass
		# make sure that the 2 initial groups (Pyro and Default) exist
		try: self.createGroup(':'+'Pyro')
		except errors.NamingError: pass
		try: self.createGroup(Pyro.config.PYRO_NS_DEFAULTGROUP)
		except errors.NamingError: pass

	def getDBDir(self):
		return self.dbroot

	def register(self,name,URI):
		self.validateName(name)
		URI=self.validateURI(URI)
		file=self.translate(name)
		try:
			self.lock.acquire()
			if os.access(file,os.R_OK):
				Log.msg('NameServer','name already exists:',name)
				raise errors.NamingError('name already exists')
			try:
				open(file,'w').write(str(URI)+'\n')
				Log.msg('NameServer','registered',name,'with URI',str(URI))
			except IOError,x:
				if x.errno==errno.ENOENT:
					raise errors.NamingError('(parent)group not found')
				elif x.errno==errno.ENOTDIR:
					raise errors.NamingError('parent is no group')
				else:
					raise errors.NamingError(x)
		finally:
			self.lock.release()

	def unregister(self,name):
		self.validateName(name)
		file=self.translate(name)
		try:
			self.lock.acquire()
			try:
				os.remove(file)
				Log.msg('NameServer','unregistered',name)
			except OSError,x:
				if x.errno==errno.ENOENT:
					raise errors.NamingError('name not found')
				elif x.errno==errno.EISDIR:
					Log.msg('NameServer','attempt to remove a group:',name)
					raise errors.NamingError('is a group')
				else:
					raise errors.NamingError(x)
		finally:
			self.lock.release()
			
	def resolve(self,name):
		# not thread-locked: higher performance and not necessary.
		self.validateName(name)
		file = self.translate(name)
		try:
			u = core.PyroURI('')
			u.reinitFromString(open(file).read())
			return u
		except IOError,x:
			if x.errno==errno.ENOENT:
				raise errors.NamingError('name not found')
			elif x.errno==errno.EISDIR:
				Log.msg('NameServer','attempt to resolve groupname:',name)
				raise errors.NamingError('attempt to resolve groupname')
			else:
				raise errors.NamingError(x)

	def flatlist(self):
		dbroot=self.translate(':')
		try:
			self.lock.acquire()
			files=self._filelist(dbroot,dbroot)
			list=[]
			for f in files:
				list.append((f, self.resolve(f)))
			return list
		finally:
			self.lock.release()

	# --- hierarchical naming support
	def createGroup(self,groupname):
		self.validateName(groupname)
		file = self.translate(groupname)
		try:
			self.lock.acquire()
			try:
				os.mkdir(file)
				Log.msg('NameServer','created group',groupname)
			except OSError,x:
				if x.errno in (errno.EEXIST, errno.EBUSY):
					raise errors.NamingError('duplicate name')
				elif x.errno == errno.ENOENT:
					raise errors.NamingError('(parent)group not found')
				else:
					raise errors.NamingError(x)
		finally:
			self.lock.release()

	def deleteGroup(self,groupname):
		self.validateName(groupname)
		if groupname==':':
			Log.msg('NameServer','attempt to deleteGroup root group')
			raise errors.NamingError('not allowed to delete root group')
		file = self.translate(groupname)
		try:
			self.lock.acquire()
			if not os.access(file,os.R_OK):
				raise errors.NamingError('group not found')
			try:
				shutil.rmtree(file)
				Log.msg('NameServer','deleted group',groupname)
			except OSError,x:
				if x.errno==errno.ENOENT:
					raise errors.NamingError('group not found')
				elif x.errno==errno.ENOTDIR:
					Log.msg('NameServer','attempt to deleteGroup a non-group:',groupname)
					raise errors.NamingError('is no group')
				else:
					raise errors.NamingError(x)
		finally:
			self.lock.release()
			
	def list(self,groupname):
		if not groupname:
			groupname=':'
		self.validateName(groupname)
		dir=self.translate(groupname)
		try:
			self.lock.acquire()
			if os.access(dir,os.R_OK):
				if os.path.isfile(dir):
					Log.msg('NameServer','attempt to list a non-group:',groupname)
					raise errors.NamingError('is no group')
				else:
					l = dircache.listdir(dir)
					list = []
					for e in l:
						if os.path.isdir(os.path.join(dir,e)):
							list.append((e,0))		# dir has code 0
						else:
							list.append((e,1))		# leaf has code 1
					return list
			raise errors.NamingError('group not found')
		finally:
			self.lock.release()

	# --- private methods follow

	# recursive file listing, output is like "find <path> -type f"
	# but using NS group separator chars
	def _filelist(self,root,path):
		try:
			(filez,dirz) = util.listdir(path)
		except OSError:
			raise errors.NamingError('group not found')
			
		files=[]
		for f in filez:
			if path==root:
				files.append(':'+f)
			else:
				p=string.replace(path[len(root):], os.sep, '.')
				files.append(':'+p+'.'+f)
		for d in dirz:
			files.extend(self._filelist(root,os.path.join(path,d)))
		return files

	# Pyro NS name to filesystem path translation
	def translate(self,name):
		if name[0]==':':
			name=name[1:]
		parts=string.split(name,'.')
		return apply(os.path.join, [self.dbroot]+parts)

		
#############################################################################
#
# The broadcast server which listens to broadcast requests of clients who
# want to discover our location, or send other system commands.
#
#############################################################################


class BroadcastServer(SocketServer.UDPServer):

	nameServerURI = ''	# the Pyro URI of the Name Server

	def server_activate(self):
		self.requestValidator=lambda x,y: 1  # default: accept all
		self.shutdown=0				# should the server loop stop?
		self.preferredTimeOut=3.0	# preferred timeout for the server loop
			
	def setNS_URI(self,URI):
		self.nameServerURI=str(URI)
	def setRequestValidator(self, validator):
		self.requestValidator=validator
	def keepRunning(self, keep):
		self.ignoreShutdown = keep	# ignore shutdown requests (i.e. keep running?)

	def bcCallback(self,ins):
		for i in ins:
			i.handle_request()

	def verify_request(self, req, addr):
		return self.requestValidator(req, addr)
	
		
class bcRequestHandler(SocketServer.BaseRequestHandler):

	def handle(self):
		Log.msg('BroadcastServer','incoming request from',str(self.client_address[0]))
		# request is a simple string
		cmd = self.request[0]
		if cmd=='location':
			# somebody wants to know our location, give them our URI
			self.request[1].sendto(self.server.nameServerURI,self.client_address)
		elif cmd=='shutdown':
			# we should die!?
			if self.server.ignoreShutdown:
				Log.msg('BroadcastServer','Shutdown ignored.')
				self.request[1].sendto('Shutdown request denied',self.client_address)
			else:
				Log.msg('BroadcastServer','Shutdown received.')
				print 'BroadcastServer received shutdown request... will shutdown shortly...'
				self.request[1].sendto('Will shut down shortly',self.client_address)
				self.server.shutdown=1
		else:
			Log.warn('BroadcastServer','Invalid command ignored:',cmd)

# The default BC request validator... accepts everything
# You must subclass this for your own validators
class BCReqValidator:
	def __call__(self, req, addr):
		(cmd,self.sock)=req
		self.addr=addr
		if cmd=='location':
			return self.acceptLocationCmd()
		elif cmd=='shutdown':
			return self.acceptShutdownCmd()
		else:
			return 0
	def reply(self,msg):
		self.sock.sendto(msg,self.addr)
	def acceptLocationCmd(self):
		return 1
	def acceptShutdownCmd(self):
		return 1


#############################################################################

class NameServerStarter:
	def __init__(self, identification=None):
		core.initServer()
		self.identification=identification
	def start(self,hstn='', nsport=0, bcport=0, keep=0, persistent=0, dbdir=None, Guards=(None,None), allowmultiple=0):
		if not nsport:
			nsport=Pyro.config.PYRO_NS_PORT
		if not bcport:
			bcport=Pyro.config.PYRO_NS_BC_PORT
		try:
			retries=Pyro.config.PYRO_BC_RETRIES
			timeout=Pyro.config.PYRO_BC_TIMEOUT
			Pyro.config.PYRO_BC_RETRIES=1
			Pyro.config.PYRO_BC_TIMEOUT=1
			try:
				uri=NameServerLocator().detectNS()
				print 'The Name Server appears to be already running on this segment.'
				print '(host:',uri.address,' port:',uri.port,')'
				if allowmultiple:
					print 'WARNING: starting another Name Server in the same segment!'
				else:
					print 'Cannot start multiple Name Servers in the same network segment.'
					return
			except errors.PyroError:
				pass
			Pyro.config.PYRO_BC_RETRIES=retries
			Pyro.config.PYRO_BC_TIMEOUT=timeout
			daemon = core.Daemon(host=hstn, port=nsport,norange=1)
		except errors.DaemonError,x:
			print 'The Name Server appears to be already running on this host.'
			print '(or somebody else occupies our port,',nsport,')'
			print 'Name Server was not started!'
			return
		if self.identification:
			daemon.setAllowedIdentifications([self.identification])
			print 'Requiring connection authentication.'
		if Guards[0]:
			daemon.setNewConnectionValidator(Guards[0])
		if persistent:
			ns=PersistentNameServer(dbdir)
			daemon.useNameServer(ns)
			NS_URI=daemon.connectPersistent(ns,constants.NAMESERVER_NAME)
		else:
			ns=NameServer()
			daemon.useNameServer(ns)
			NS_URI=daemon.connect(ns,constants.NAMESERVER_NAME)
	
		try:
			self.bcserver = BroadcastServer((hstn or '',bcport),bcRequestHandler)
		except socket.error:
			print 'Somebody else is occupying our broadcast port,',bcport
			print 'Name Server was not started!'
			return
		if Guards[1]:
			self.bcserver.setRequestValidator(Guards[1])
		self.bcserver.keepRunning(keep)
		if keep:
			print 'Will ignore shutdown requests.'
		else:
			print 'Will accept shutdown requests.'

		if Guards[0] or Guards[1]:
			print 'Using security plugins:'
			if Guards[0]:
				clazz=Guards[0].__class__
				print '  NS new conn validator =',clazz.__name__,'from', clazz.__module__, ' ['+sys.modules.get(clazz.__module__).__file__+']'
			else: print '  default NS new conn validator'
			if Guards[1]:
				clazz=Guards[1].__class__
				print '  BC request validator  =',clazz.__name__,'from', clazz.__module__, ' ['+sys.modules.get(clazz.__module__).__file__+']'
			else: print '  default BC request validator'

		ns.publishURI(NS_URI)
		self.bcserver.setNS_URI(NS_URI)
		Log.msg('NS daemon','This is the Pyro Name Server.')
		if persistent:
			Log.msg('NS daemon','Persistent mode, database is in',ns.getDBDir())
			print 'Persistent mode, database is in',ns.getDBDir()
		Log.msg('NS daemon','Starting on',daemon.hostname,'port',
		  daemon.port, ' broadcast server on port',bcport)

		print 'Name Server started.'

		# I use a timeout here otherwise you can't break gracefully on Windoze
		try:
			daemon.setTimeout(20)  # XXX fixed timeout
			daemon.requestLoop(lambda s=self: not s.bcserver.shutdown,
				self.bcserver.preferredTimeOut,[self.bcserver],self.bcserver.bcCallback)
			if self.bcserver.shutdown:
				self.shutdown(ns)
		except KeyboardInterrupt:
			Log.warn('NS daemon','shutdown on user break signal')
			print 'Shutting down on user break signal.'
			self.shutdown(ns)
		except:
			try:
				import traceback
				(exc_type, exc_value, exc_trb) = sys.exc_info()
				out = ''.join(traceback.format_exception(exc_type, exc_value, exc_trb)[-5:])
				Log.error('NS daemon', 'Unexpected exception, type',exc_type,
					'\n--- partial traceback of this exception follows:\n',
					out,'\n--- end of traceback')
				print '*** Exception occured!!! Partial traceback:'
				print out
				print '*** Resuming operations...'
			finally:	
				del exc_type, exc_value, exc_trb

		Log.msg('NS daemon','Shut down gracefully.')
		print 'Name Server gracefully stopped.'

	def shutdown(self, ns):
		daemon=ns.getDaemon()
		daemon.disconnect(ns) # clean up nicely
		self.bcserver.shutdown=1
		daemon.shutdown()


def main():
	if __name__!='__main__':
		del sys.argv[0]	# remove '-c' arg
	Args = util.ArgParser()
	Args.parse(sys.argv,'hkmn:p:b:d:s:i:')
	try:
		Args.getOpt('h')
		print 'Usage: ns [-h] [-k] [-m] [-n hostname] [-p port] [-b port] [-i identification]'
		print '          [-d [databasefile]] [-s securitymodule]'
		print '  where -p = NS server port'
		print '        -b = NS broadcast port'
		print '        -n = non-default hostname to bind on'
		print '        -m = allow multiple instances in network segment'
		print '        -k = keep running- do not respond to shutdown requests'
		print '        -d = use persistent database, provide optional storage directory'
		print '        -s = use given python module with security plugins'
		print '        -i = specify the required authentication ID'
		print '        -h = print this help'
		raise SystemExit
	except KeyError:
		pass
	host = Args.getOpt('n','')
	port = int(Args.getOpt('p',Pyro.config.PYRO_NS_PORT))
	bcport = int(Args.getOpt('b',Pyro.config.PYRO_NS_BC_PORT))
	ident = Args.getOpt('i',None)

	try:
		dbdir = Args.getOpt('d')
		persistent = 1
	except KeyError:
		persistent = 0
		dbdir = None

	try:
		Args.getOpt('k')
		keep=1
	except KeyError:
		keep=0
	try:
		Args.getOpt('m')
		allowmultiple=1
	except KeyError:
		allowmultiple=0

	try:
		secmod = __import__(Args.getOpt('s'),locals(),globals())
		Guards = (secmod.NSGuard(), secmod.BCGuard())
	except ImportError,x:
		print 'Error loading security module:',x
		print '(is it in your python import path?)'
		raise SystemExit
	except KeyError:
		secmod = None
		Guards = (None,None)

	Args.printIgnored()
	if Args.args:
		print 'Ignored arguments:',string.join(Args.args)

	print '*** Pyro Name Server ***'
	if ident:
		starter=NameServerStarter(identification=ident)
	else:	
		starter=NameServerStarter()

	starter.start(host,port,bcport,keep,persistent,dbdir,Guards,allowmultiple)

